/************************************************************
* File: studentlistfunc.c
* This file contains function definitions for the GPA library
* using a linked list.
************************************************************/
#include <stdio.h>
#include <string.h>       /* for strcmp() and strdup() */
#include <stdlib.h>       /* for malloc() and free() */
#include "studentfile.h"  /* header file GPA lib */

/* line buffer size */
#define BUFSIZE 1024

#define TRUE 1
#define FALSE 0

/* static function in file scope */
static
StudentNode highestLowestGPA(StudentNode head, int isFindHighest);


/* external functions */

/* Code block 1: functions for a single student */
void gpa_printStudent(StudentNode sn) {
    if(sn == NULL) {
        fprintf(stderr, "Error: invalid argument passed to printStudent()\n");
        return;
    }
    printf("  id: %d; name: %6s; birthday: %8s %2d, %d; GPA: %.2f\n",
           sn->id, sn->name, gpa_monthString(sn->birthday.month),
           sn->birthday.day, sn->birthday.year, sn->gpa);
}      

int gpa_getId(StudentNode sn) {
    return sn->id;
}

char *gpa_getName(StudentNode sn) {
    return sn->name;
}

struct Birthday gpa_getBirthday(StudentNode sn) {
    return sn->birthday;
}

double gpa_getGPA(StudentNode sn) {
    return sn->gpa;
}

void gpa_setGPA(StudentNode sn, double gpa) {
    sn->gpa = gpa;
}

/* create a student node */
StudentNode gpa_createStudent(int id, const char *name, 
                     struct Birthday b, double gpa) {
    StudentNode sn;
  
    sn = (StudentNode) malloc(sizeof(struct ListStudent));
    if(sn == NULL) {
        fprintf(stderr, "Error: not enough memory.\n");
        return NULL;
    }
    sn->id = id;
    sn->name = strdup(name);
    sn->birthday = b;
    sn->gpa = gpa;
    sn->next = NULL;
    return sn;
}

/* free the memory for a student node and related memory */
void gpa_freeStudent(StudentNode sn) {
    if(sn) {
        if(sn->name) {
            free(sn->name);
        }
        free(sn);
    }
}
  

/* Code block 2: functions for a list of students */
/* get total number of students in the list */
int gpa_totnum(StudentNode head) {
    int totnum = 0;

    while(head != NULL) {
        totnum++;
        head = head->next;
    }
    return totnum;
}

/* print info for all students in the list */
void gpa_printList(StudentNode head) {
    while(head != NULL) {
        gpa_printStudent(head);
        head = head->next;
    }
}   

/* calculate the average of the GPA for students in the list */
double gpa_averageGPA(StudentNode head) {
    double gpa, totalval;
    int i;
 
    totalval=0.0;
    for(i=0; head != NULL; head = head->next, i++) { 
         totalval += head->gpa;
    }
    gpa = totalval/i;
    return gpa;
}

/* find the student with the highest GPA in the list */
StudentNode gpa_highestGPA(StudentNode head) {
    return highestLowestGPA(head, TRUE);
}
  
/* find the student with the lowest GPA in the list */
StudentNode gpa_lowestGPA(StudentNode head) {
    return highestLowestGPA(head, FALSE);
}

/* find the student based on the id and name */
StudentNode gpa_find(StudentNode head, int id, const char *name) {
    while(head!=NULL && !(!strcmp(head->name, name) && 
          head->id == id)) {
       head = head->next;
    }
    return head;
}

/* print out the students with GAP above 'gpa' */
int gpa_aboveGPA(StudentNode head, double gpa) {
    int num;
    for(num=0; head != NULL; head=head->next) { 
        if(head->gpa >= gpa) {
            gpa_printStudent(head);
            num++;
        }
    }
    return num;
}

/* sort students with GPA in descending order using selection sort */
void gpa_sortGPA(StudentNode head) {
    struct ListStudent tmp;           /* for contents of a node */
    StudentNode next, sni, snj, maxi; /* temporary nodes */

    sni = head; /* start from the beginning for the outer loop */
    snj = head; /* start from the beginning for the inner loop */
    while(sni!= NULL) {
        maxi = sni;        /* assume sni is the maximum */
        while(snj!= NULL) {
            if(maxi->gpa < snj->gpa) {
                maxi = snj; /* save the node for the new maximum */
            }
            snj = snj->next;  /* process the next node */
        }
        /* swap contents of maxi and sni, update next links */
        if(maxi != sni) {       
            tmp = *sni;         /* save sni */
            next = sni->next;   /* save pointer next for node sni */
            *sni = *maxi;       /* sni is set with the value for maxi */
            sni->next = next;   /* update pointer next for node sni */
            next = maxi->next;  /* save pointer next for node maxi */
            *maxi = tmp;        /* maxi has the value of sni */
            maxi->next = next;  /* update pointer next for node maxi */
        }
        /* move to the next node */
        sni = sni->next;
        snj = sni;
    }
}

/* read student data file and return the data in list with the allocated memory */
StudentNode gpa_readData(const char *filename) {
    StudentNode head, sn; /* declare head of the list of students,
                             and a single student */
    FILE *stream;         /* a file stream for reading data from filename */
    char line[BUFSIZE];   /* a buffer to hold the a line from filename */
    int id;                   /* student id */
    char name[32];            /* student name */
    struct Birthday birthday; /* student birthday */
    double gpa;               /* student GPA */

    stream = fopen(filename, "r");  /* open filename for reading */
    if(stream == NULL) {
        fprintf(stderr, "Error: cannot open file '%s' for reading\n", filename);
        return NULL;
    }

    head = NULL;
    fgets(line, BUFSIZE, stream); /* read a line into the buffer */
    while(!feof(stream)) {
        while(line[0] == '#') { /* skip the comment */
           fgets(line, BUFSIZE, stream); /* read a line into the buffer */
        }
       /* read the data for a student from string line and save them in
           the memory pointed by sn, id and name */
        sscanf(line, "%d%s%hd%hd%hd%lf", &id, name,
               &birthday.day, &birthday.month, &birthday.year, &gpa);
        /* create a node with id, name, birthday, and gpa */
        sn = gpa_createStudent(id, name, birthday, gpa);
        gpa_append(&head, sn);        /* append the node into the list */
        fgets(line, BUFSIZE, stream); /* read a line into the buffer */
    }
    fclose(stream);         /* close the file stream */
    return head;            /* return the head of the list */
}

/* write student data into a file */
int gpa_writeData(StudentNode head, const char *filename) {
    FILE *stream;     /* a file stream for writing data int filename */

    stream = fopen(filename, "w");  /* open filename for writing */
    if(stream == NULL) {
        fprintf(stderr, "Error: cannot open file '%s' for writing\n", filename);
        return -1;
    }
    while(head != NULL) {  /* write until the sentinel value for the last link */
        fprintf(stream, "%d %s %hd %hd %hd %g\n", head->id, head->name,
               head->birthday.day, head->birthday.month, head->birthday.year,
               head->gpa);
        head = head->next; /* advance to the next element */
    }
    fclose(stream);        /* close the file stream */
    return 0;
}

/* delete all nodes and their associated memory of the list */
void gpa_clearList(StudentNode head) {
    StudentNode sn;

    while(head) {
       sn = head;
       head = head->next;
       gpa_freeStudent(sn);
    }
}

/* prepend to the list at the beginning */
void gpa_prepend(StudentNode *head, StudentNode sn) {
    sn->next = *head;/* head becomes the next node of the new node */
    *head = sn;      /* the new node becomes the new head */
}

/* append to the list at end */
void gpa_append(StudentNode *head, StudentNode sn) {
    StudentNode tmp = *head; /* tmp is the head of the list */

    if(*head == NULL) /* no node in linked list */
      *head = sn;
    else { 
       /* get the tail of the linked list */ 
       while(tmp->next) 
          tmp = tmp->next;
       /* add the new one to the end of the list */
       tmp->next = sn;
    }
} 

/* insert to the list at pos */
int gpa_insert(StudentNode *head, StudentNode sn, int pos) {
    StudentNode tmp = *head; /* tmp is the head of the list */
    int i = 0;

    /* insert to the first position */
    if(pos == 1) {
        *head = sn;
        (*head)->next = tmp;
        return 0;
    }

    /* get position */
    while(tmp) {
        if(++i >= pos-1) /* match or invalid position */
            break;
        tmp = tmp->next;
    }
    if(i != pos-1) { /* can not get that position */
        fprintf(stderr, "Error: Wrong position for insert().\n");
        return -1;
    }
    /* update links */
    sn->next = tmp->next;
    tmp->next = sn;
    return 0;
}  

/* remove a node from the list */
void gpa_remove(StudentNode *head, StudentNode sn) {
    StudentNode tmp = *head; /* tmp is the head of the list */

    if (*head == NULL)       /* empty node */
        return;
    else if ((*head) == sn) {  /* sn is the head */
        *head = (*head)->next;
        gpa_freeStudent(tmp);  /* free the memory for the node */

    }
    else {
        while (tmp->next) {
            if (tmp->next == sn) {  /* found the node */
                tmp->next = tmp->next->next; /* detach the node */
                gpa_freeStudent(sn); /* free the memory for the node */
                return; 
             }
             tmp = tmp->next;
         }
    }
}  


/* Code block 3: utility functions */
const char *gpa_monthString(int month) {
    static const char *monthName[] = {"January", "February", "March", 
           "April", "May", "June", "July", "August", "September", 
           "October", "November", "December"};

    return (month < 1 || month > 12) ? NULL : monthName[month-1];
}


/* Code block 4: internal functions */

/* find the student with the highest or lowest GPA in the list */
static
StudentNode highestLowestGPA(StudentNode head, int isFindHighest) {
    StudentNode sn;
    double gpa;

    gpa = head->gpa;       /* initial gpa */
    sn =  head;            /* initial sn */
    while(head != NULL) {
        if(isFindHighest == TRUE && head->gpa > gpa) {
            gpa = head->gpa;
            sn= head;
        }
        else if(isFindHighest == FALSE && head->gpa < gpa) {
            gpa = head->gpa;
            sn = head;
        }
        head = head->next;
    }

    return sn;
}
