/*****************************************************
 *  File: dlistlib.c
 *  Definitions of functions for doubly linked lists 
 ****************************************************/
#include <stdio.h>
#include <string.h> /* for strdup() and strcmp() */
#include <stdlib.h> /* for malloc() and free() */
#include "dlist.h"  /* header file for doubly linked lists */

/* create a node of a list */
Node createNode(int id, const char *name) {
    Node nd;

    nd = (Node) malloc(sizeof(ListNode));
    if(nd == NULL) {
        fprintf(stderr, "Error: not enough memory.\n");
        return NULL;
    }
    nd->id = id;            /* set member id */
    nd->name = strdup(name);/* set member name */
    nd->next = NULL;        /* set member next, point to NULL */
    nd->prev = NULL;        /* set member prev, point to NULL */
    return nd;
}

/* freen a node in a list */
void freeNode(Node nd) {
    if(nd) {
        if(nd->name) {      /* if name is not NULL */
            free(nd->name);
        }
        free(nd);
    }
}

/* prepend to the list at the beginning */
void prepend(Node *head, Node nd) {
    nd->next = *head;/* head becomes the next node of the new node */
    if(*head) {      /* the new node becomes head's previous node */
      (*head)->prev = nd; 
    }
    *head = nd;      /* the new node becomes the new head */
}

/* append to the list at end */
void append(Node *head, Node nd) {
    Node tmp = *head; /* tmp is the head of the list */

    if(*head == NULL) /* no node in linked list */
        *head = nd;
    else {
       /* get the tail of the linked list */
       while(tmp->next)
          tmp = tmp->next;
       /* add the new one to the end of the list */
       tmp->next = nd; /* new nd becomes the last one */
       nd->prev = tmp; /* update the link to the previous node */
    }
} 

/* insert to the list at pos */
int insert(Node *head, Node nd, int pos) {
    int i = 0;
    Node tmp = *head; /* tmp is the head of the list */

    /* insert to the first position */
    if(pos == 1) {
       *head = nd;
       nd->next = tmp;
       tmp->prev = nd;
       return 0;
    }

    /* get position */
    while(tmp) {
      if(++i >= pos-1) /* match or invalid position */
        break;
      tmp = tmp->next;
    }
    if(i != pos-1) { /* can not get that position */
      fprintf(stderr, "Error: Wrong position for insert(). \n");
      return -1;
    }
    /* update links */
    nd->next = tmp->next;
    if(tmp->next != NULL)   /* not insert into the last node */
      tmp->next->prev = nd;
    tmp->next = nd;
    nd->prev = tmp;
    return 0;
}     
   
/* find the node */
Node find(Node head, int id, const char *name) {
    while(head!=NULL && 
       !(head->id == id && !strcmp(head->name, name))) {
       head = head->next;
    }
    return head;
}

/* remove a node from the list */
void remov(Node *head, Node nd) {
    Node tmp = *head; /* tmp is the head of the list */

    if (*head == NULL)       /* empty list */
      return;
    else if (*head == nd) {  /* nd is the head */
      *head = (*head)->next;
      (*head)->prev = NULL;
      freeNode(nd);          /* free the memory for the node */
    }
    else {
      while (tmp->next) {
        if (tmp->next == nd) {  /* found the node */
          tmp->next = tmp->next->next; /* detach the node */
          if(tmp->next != NULL) /* not remove the last node */
            tmp->next->prev = tmp;     /* detach the node */
          freeNode(nd);      /* free the memory for the node */
          return;
        }
        tmp = tmp->next;
      }
    }
}   

/* print id and name in a node */
void print(Node nd) {
    if(nd) {
      printf("id  = %d, name = %s\n", nd->id, nd->name);
    }
}

/* print id and name in each node of the list */
void printList(Node head) {
    while(head) {
      print(head);
      head = head->next;
    }
}

/* delete all nodes and their associated memory of the list */
void clearList(Node head) {
    Node nd;

    while(head) {
      nd = head;
      head = head->next;
      freeNode(nd);
    }
}
